using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Phidgets;
using System.Threading;

namespace MacanumRobot
{
    public partial class Form1 : Form
    {
        const string robotServerID = "mecanum";
        //const string robotServerID = "phidgetsbc";

        const string frontRightLabel = "FrontRight";
        const string frontLeftLabel = "FrontLeft";
        const string rearRightLabel = "RearLeft";
        const string rearLeftLabel = "RearRight";

        const int joyStickXAxisIndex = 0;
        const int joyStickYAxisIndex = 1;
        const int joyStickZAxisIndex = 2;

        //For any motors which turn backwards.
        const int frontRightDir = 1;
        const int frontLeftDir = -1;
        const int rearRightDir = 1;
        const int rearLeftDir = -1;

        const string webcamAddress = "mecanum.local";
        //const string webcamAddress = "phidgetsbc.local";
        const string webcamPort = "81";

        const int setMotorDelay = 50; //in ms - so 20 updates/second




        double heading = 0, headingLockPoint = 0;
        int xMid = 2048, yMid = 2048, zMid = 2048;

        InterfaceKit controller;
        Dictionary dict;
        KeyListener listener;

        double m_maxOutput = 100.0;

        public Form1()
        {
            //Phidget.enableLogging(Phidget.LogLevel.PHIDGET_LOG_INFO, null);
            InitializeComponent();
            setMotorsThread = new Thread(new ThreadStart(setMotorsThreadFunction));
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            controller = new InterfaceKit();

            setMotorsThread.Start();

            controller.Attach += new Phidgets.Events.AttachEventHandler(controller_Attach);
            controller.Detach += new Phidgets.Events.DetachEventHandler(Detach);
            controller.SensorChange += new Phidgets.Events.SensorChangeEventHandler(controller_SensorChange);
            controller.open();


            dict = new Dictionary();
            listener = new KeyListener(dict, "^/robot/stat/.*");

            listener.KeyChange += new Phidgets.Events.KeyEventHandler(listener_KeyChange);
            listener.KeyRemoval += new Phidgets.Events.KeyEventHandler(listener_KeyRemoval);

            dict.ServerConnect += new Phidgets.Events.ServerConnectEventHandler(dict_ServerConnect);
            dict.ServerDisconnect += new Phidgets.Events.ServerDisconnectEventHandler(dict_ServerDisconnect);
            dict.open(robotServerID);
        }

        int remoteAttachCount = 0;

        void listener_KeyRemoval(object sender, Phidgets.Events.KeyEventArgs e)
        {
            switch (e.key)
            {
                case "/robot/stat/attachCount":
                    remoteAttachCount = 0;
                    statusTxt.Text = attachCount.ToString() + "/5 Attached";
                    break;
                default:
                    break;
            }
        }

        void listener_KeyChange(object sender, Phidgets.Events.KeyEventArgs e)
        {
            double speed = 0;
            switch (e.key)
            {
                case "/robot/stat/attachCount":
                    remoteAttachCount = int.Parse(e.value);
                    statusTxt.Text = attachCount.ToString() + "/5 Attached";
                    break;
                case "/robot/stat/MotorSpeeds":
                    string[] speeds = e.value.Split(new char[] { ',' });
                    speed = double.Parse(speeds[0]);
                    frontLeftTxt.Text = (speed * frontLeftDir).ToString("F1") + "%";
                    speed = double.Parse(speeds[1]);
                    frontRightTxt.Text = (speed * frontRightDir).ToString("F1") + "%";
                    speed = double.Parse(speeds[2]);
                    rearLeftTxt.Text = (speed * rearLeftDir).ToString("F1") + "%";
                    speed = double.Parse(speeds[3]);
                    rearRightTxt.Text = (speed * rearRightDir).ToString("F1") + "%";
                    break;
                case "/robot/stat/MotorAccels":
                    accelTxt.Text = e.value;
                    break;
                case "/robot/stat/heading":

                    double tempheading = double.Parse(e.value);
                    while (tempheading < 0)
                        tempheading += 360;
                    while (tempheading > 360)
                        tempheading -= 360;
                    heading = tempheading;

                    headingTxt.Text = heading.ToString("F1") + "";

                    break;
                default:
                    break;
            }
        }

        void dict_ServerDisconnect(object sender, Phidgets.Events.ServerDisconnectEventArgs e)
        {
            listener.stop();
        }

        void dict_ServerConnect(object sender, Phidgets.Events.ServerConnectEventArgs e)
        {
            listener.start();
        }

        void Detach(object sender, Phidgets.Events.DetachEventArgs e)
        {
            statusTxt.Text = attachCount.ToString() + "/5 Attached";
        }

        void controller_Attach(object sender, Phidgets.Events.AttachEventArgs e)
        {
            controller.sensors[joyStickXAxisIndex].Sensitivity = 1;
            controller.sensors[joyStickYAxisIndex].Sensitivity = 1;
            controller.sensors[joyStickZAxisIndex].Sensitivity = 1;

            xMid = controller.sensors[joyStickXAxisIndex].RawValue;
            yMid = controller.sensors[joyStickYAxisIndex].RawValue;
            zMid = controller.sensors[joyStickZAxisIndex].RawValue;

            statusTxt.Text = attachCount.ToString() + "/5 Attached";
        }

        int attachCount
        {
            get
            {
                int count = remoteAttachCount;
                if (controller.Attached)
                    count++;
                return count;
            }
        }

        void controller_SensorChange(object sender, Phidgets.Events.SensorChangeEventArgs e)
        {
            JoystickDrive((controller.sensors[joyStickXAxisIndex].RawValue - xMid) / 2000.0,
                (controller.sensors[joyStickYAxisIndex].RawValue - yMid) / 2000.0,
                -(controller.sensors[joyStickZAxisIndex].RawValue - zMid) / 1500.0);
        }

        double Limit(double val)
        {
            if (val > 1.0)
                return 1.0;
            if (val < -1.0)
                return -1.0;
            return val;
        }

        void Normalize(double[] vals)
        {
            double max = 0;
            foreach (double val in vals)
                if (Math.Abs(val) > max)
                    max = Math.Abs(val);

            if (max > 1.0)
                for (int i = 0; i < vals.Length; i++)
                    vals[i] /= max;
        }

        /* values should be -1 - 1
         * */
        void JoystickDrive(double x, double y, double z)
        {

            //Dead Space
            double xyDeadSpace = 0.1;
            double zDeadSpace = 0.25;

            double newx = (Math.Abs(x) - xyDeadSpace) * (xyDeadSpace+1);
            if (newx < 0) newx = 0;
            x = (x < 0) ? -newx : newx;

            double newy = (Math.Abs(y) - xyDeadSpace) * (xyDeadSpace + 1);
            if (newy < 0) newy = 0;
            y = (y < 0) ? -newy : newy;

            double newz = (Math.Abs(z) - zDeadSpace) * (zDeadSpace + 1);
            if (newz < 0) newz = 0;
            z = (z < 0) ? -newz : newz;

            x = Limit(x);
            y = Limit(y);
            z = Limit(z);

            double magnitude = Math.Sqrt(x * x + y * y);
            double direction = Math.Atan2(x, y);
            double rotation = z;

            if (magnitude == 0)
                direction = 0;

            //Into degrees
            direction = direction * 180.0 / Math.PI;

            if (gyroLocked)
            {
                direction -= heading;
                direction += headingLockPoint;
            }

            if (direction < 0)
                direction += 360;
            if (direction > 360)
                direction -= 360;

            MecanumDrive(magnitude, direction, rotation);
        }

        /*
         * Magnitude is 0-1
         * Direction is 0-360
         * Rotation is 0-1
         * */
        void MecanumDrive(double magnitude, double direction, double rotation)
        {
	        //Limit limits magnitude to 1.0
	        magnitude = Limit(magnitude);

            velocityTxt.Text = (magnitude * m_maxOutput).ToString("F1") + "%";
            dirTxt.Text = direction.ToString("F0");
            yawTxt.Text = (rotation * m_maxOutput).ToString("F1") + "%";

            // Normalized for full power along the Cartesian axes.
            magnitude = magnitude * Math.Sqrt(2.0);
        	
	        // The rollers are at 45 degree angles.
	        double dirInRad = (direction + 45.0) * Math.PI / 180.0;
	        double cosD = Math.Cos(dirInRad);
	        double sinD = Math.Sin(dirInRad);
            double[] wheelSpeeds = new double[4];
        	
	        wheelSpeeds[0] = sinD * magnitude + rotation;
	        wheelSpeeds[1] = cosD * magnitude - rotation;
	        wheelSpeeds[2] = cosD * magnitude + rotation;
	        wheelSpeeds[3] = sinD * magnitude - rotation;
        	
	        //Only if any are > 1.0
	        Normalize(wheelSpeeds);

            frontLeftSpeed = (wheelSpeeds[0] * m_maxOutput * frontLeftDir);
            frontRightSpeed = (wheelSpeeds[1] * m_maxOutput * frontRightDir);
            rearLeftSpeed  = (wheelSpeeds[2] * m_maxOutput * rearLeftDir);
            rearRightSpeed = (wheelSpeeds[3] * m_maxOutput * rearRightDir);

            //frontLeftTxt.Text = (frontLeftSpeed * frontLeftDir).ToString("F1") + "%";
            //frontRightTxt.Text = (frontRightSpeed * frontRightDir).ToString("F1") + "%";
            //rearLeftTxt.Text = (rearLeftSpeed * rearLeftDir).ToString("F1") + "%";
            //rearRightTxt.Text = (rearRightSpeed * rearRightDir).ToString("F1") + "%";
        }

        Thread setMotorsThread;
        Boolean settingMotors = true;
        void setMotorsThreadFunction()
        {
            while (settingMotors)
            {
                if (attachCount == 5)
                    setMotors();
                Thread.Sleep(setMotorDelay);
            }
        }

        double frontLeftSpeed = 0, frontRightSpeed = 0, rearLeftSpeed = 0, rearRightSpeed = 0;
        void setMotors()
        {

            try
            {
                dict.add("/robot/setMotorSpeeds", frontLeftSpeed.ToString()+","+
                    frontRightSpeed.ToString()+","+
                rearLeftSpeed.ToString() + "," +
                rearRightSpeed.ToString());
            }
            catch { }
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            settingMotors = false;

            //Stop the robot!
            frontLeftSpeed = frontRightSpeed = rearLeftSpeed = rearRightSpeed = 0;
            setMotors();

            controller.Attach -= controller_Attach;
            controller.Detach -= Detach;

            controller.SensorChange -= controller_SensorChange;

            listener.KeyChange -= listener_KeyChange;
            listener.KeyRemoval -= listener_KeyRemoval;

            dict.ServerConnect -= dict_ServerConnect;
            dict.ServerDisconnect -= dict_ServerDisconnect;

            Application.DoEvents();
            setMotorsThread.Join(250);

            try
            {
                if (webcam != null && webcam.Visible)
                    webcam.stop();
            }
            catch { }

            listener.stop();
            dict.close();

            controller.close();
        }

        WebcamViewer webcam = null;
        private void webcamBtn_Click(object sender, EventArgs e)
        {
            webcam = new WebcamViewer(webcamAddress, webcamPort);
            webcam.Show();
        }

        private void trackBar1_Scroll(object sender, EventArgs e)
        {
            try
            {
                dict.add("/robot/setMotorAccels", trackBar1.Value.ToString() );
            }
            catch { }
        }

        Boolean gyroLocked = false;
        private void lockGyroButton_Click(object sender, EventArgs e)
        {
            if (gyroLocked)
            {
                gyroLocked = false;
                lockGyroButton.Text = "Lock";
            }
            else
            {
                headingLockPoint = heading;
                gyroLocked = true;
                lockGyroButton.Text = "Un-Lock";
            }
        }
    }
}