using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Drawing;
using System.Net;
using System.IO;
using System.ComponentModel;

namespace MJPEGDecoder
{
    public class MJPEGDecodeAndDisplay
    {
        private PictureBox display = null;
        private String address, pass, user;

        private bool stopStream = false;

        private HttpWebRequest req;
        private Stream imageStream = null;
        private Byte[] boundary;

        private Thread streamThread = null;
        private bool streamState = false;

        public bool isStreaming
        {
            get
            {
                return streamState;
            }
        }

        public MJPEGDecodeAndDisplay(String address)
        {
            //this.display = displayWindow;
            this.address = address;
            this.user = null;
            this.pass = null;
        }

        public MJPEGDecodeAndDisplay(String address, String username, String password)
        {
            //this.display = displayWindow;
            this.address = address;
            this.user = username;
            this.pass = password;
        }

        ~MJPEGDecodeAndDisplay()
        {
            stop();
        }

        public class PictureEventArgs : EventArgs
        {
            public readonly Image picture;
            public PictureEventArgs(Image picture)
            {
                this.picture = picture;
            }
        }
        public delegate void PictureEventHandler(object sender, PictureEventArgs e);
        public event PictureEventHandler Picture;
        private void OnPicture(PictureEventArgs e)
        {
            if (null != Picture)
            {
                foreach (PictureEventHandler PictureHandler in Picture.GetInvocationList())
                {
                    ISynchronizeInvoke syncInvoke = PictureHandler.Target as ISynchronizeInvoke;
                    if ((null != syncInvoke) && (syncInvoke.InvokeRequired))
                    {
                        syncInvoke.Invoke(PictureHandler, new object[] { this, e });
                    }
                    else
                        PictureHandler(this, e);
                }
            }
        }

        public class FPSEventArgs : EventArgs
        {
            public readonly double fps;
            public FPSEventArgs(double fps)
            {
                this.fps = fps;
            }
        }
        public delegate void FPSEventHandler(object sender, FPSEventArgs e);
        public event FPSEventHandler FPS;
        private void OnFPS(FPSEventArgs e)
        {
            if (null != FPS)
            {
                foreach (FPSEventHandler FPSHandler in FPS.GetInvocationList())
                {
                    ISynchronizeInvoke syncInvoke = FPSHandler.Target as ISynchronizeInvoke;
                    if ((null != syncInvoke) && (syncInvoke.InvokeRequired))
                    {
                        syncInvoke.Invoke(FPSHandler, new object[] { this, e });
                    }
                    else
                        FPSHandler(this, e);
                }
            }
        }

        public void start()
        {
            if (streamState == true)
                throw new Exception("Streaming is already active");

            stop();

            //make an initial connection so we know if it works
            HttpWebResponse resp = null;
            try
            {
                req = (HttpWebRequest)WebRequest.Create("http://" + address);
                if (user != null && pass != null)
                    req.Credentials = new NetworkCredential(user, pass);
                resp = (HttpWebResponse)req.GetResponse();
                ASCIIEncoding encoding = new ASCIIEncoding();
				boundary = encoding.GetBytes(resp.ContentType.Substring(resp.ContentType.IndexOf("boundary=") + 9));
                imageStream = resp.GetResponseStream();

                if (resp.ContentType.IndexOf("multipart/x-mixed-replace") == -1)
                    throw new Exception("Invalid URL");
            }
            catch (Exception ex)
            {
                if (imageStream != null)
                    imageStream.Close();
                imageStream = null;
                throw ex;
            }
            finally
            {
                //if(resp!=null)
                  //  resp.Close();
            }

            stopStream = false;
            streamThread = new Thread(new ThreadStart(streamingThreadFunction));
            streamThread.Start();

            streamState = true;
        }

        public void stop()
        {
            if (streamThread != null)
            {
                if (streamThread.IsAlive == true)
                {
                    stopStream = true;
                    if(streamThread.Join(1000) == false)
                        streamThread.Abort();
                }
            }
            streamState = false;
            streamThread = null;
        }

        private delegate void PictureBoxDelegate(PictureBox pic, Image img);
        private void setPicture(PictureBox pic, Image img)
        {
            if (pic.InvokeRequired)
                try { pic.Invoke(new PictureBoxDelegate(setPicture), new Object[] { pic, img }); }
                catch { }
            else
            {
                pic.Image = img;
            }
        }

        private bool byteArrayMatch(Byte[] matchme, Byte[] inme, int offset)
        {
            Boolean match = true;
            for(int i = 0;i<matchme.Length;i++)
                if (matchme[i] != inme[i + offset])
                {
                    match = false;
                    break;
                }
            return match;
        }

        private void streamingThreadFunction()
        {
            try
            {
                //buffer needs to be big enough for two images
                Byte[] buffer = new byte[100000];
                int readoffset = 0, writeoffset=0, boundaryOffset = 0, lastBoundaryOffset = 0;
                int bytesRead;
                bool boundaryFound = false;
                bool imageFound = false;
                int imageCount = 0;
                Byte[] delimiter = new byte[4] { 13, 10, 13, 10 };

                DateTime start = DateTime.Now;

                while (true)
                {
                    if (stopStream)
                        return;
                    //stream to pictureBox
                    try
                    {
                        bytesRead = imageStream.Read(buffer, writeoffset, buffer.Length - writeoffset);
                        writeoffset += bytesRead;

                        //go through all read data
                        if (bytesRead >= boundary.Length)
                        {
                            while (true)
                            {
                                if (stopStream)
                                    return;
                                //look for the boundary
                                bool boundaryNow = false;
                                int i = 0;
                                for (i = readoffset; i <= writeoffset - boundary.Length; i++)
                                {
                                    if (stopStream)
                                        return;
                                    readoffset++;
                                    if (byteArrayMatch(boundary, buffer, i))
                                    {
                                        if (boundaryFound)
                                            imageFound = true;
                                        boundaryFound = true;
                                        lastBoundaryOffset = boundaryOffset;
                                        boundaryOffset = i;
                                        boundaryNow = true;
                                        readoffset += boundary.Length;
                                        break;
                                    }
                                }

                                //go read some more
                                if (!boundaryNow)
                                    break;

                                if (imageFound)
                                {
                                    //parse out the image, display it
                                    bool foundImageStart = false;
                                    int imageStart = 0;
                                    for (int k = lastBoundaryOffset; k <= boundaryOffset; k++)
                                    {
                                        if (stopStream)
                                            return;
                                        if (byteArrayMatch(delimiter, buffer, k))
                                        {
                                            foundImageStart = true;
                                            imageStart = k+4;
                                            break;
                                        }
                                    }
                                    if (foundImageStart)
                                    {
                                        if (stopStream)
                                            return;
                                        try
                                        {
                                            Image img = Image.FromStream(new MemoryStream(buffer, imageStart, boundaryOffset-imageStart));

                                            //Console.WriteLine("Image: " + img.Width.ToString() + " x " + img.Height.ToString());

                                            // Display
                                            if(display != null)
                                                setPicture(display, img);
                                            //Event
                                            OnPicture(new PictureEventArgs(img));
                                        }
                                        catch (ThreadAbortException)
                                        {
                                            //aborted - return
                                            return;
                                        }
                                        catch (Exception ex)
                                        {
                                            System.Diagnostics.Debug.WriteLine("=============: " + ex.Message);
                                        }
                                    }
                                    else
                                    {
                                        System.Diagnostics.Debug.WriteLine("Trouble parsing out an image - image buffer size = " + (boundaryOffset - lastBoundaryOffset).ToString());
                                    }
                                    
                                    //make more room in the buffer
                                    int j = 0;
                                    for (int k = boundaryOffset; k < writeoffset; k++)
                                    {
                                        if (stopStream)
                                            return;
                                        buffer[j] = buffer[k];
                                        j++;
                                    }
                                    readoffset -= boundaryOffset;
                                    writeoffset -= boundaryOffset;
                                    boundaryOffset -= boundaryOffset;

                                    imageFound = false;
                                    imageCount++;

                                    TimeSpan time = DateTime.Now.Subtract(start);

                                    if (time.TotalSeconds >= 1)
                                    {
                                        OnFPS(new FPSEventArgs(imageCount));
                                        imageCount = 0;
                                        start = start.AddSeconds(1);
                                    }
                                    //Console.WriteLine("Image " + imageCount.ToString());
                                }

                            }
                        }
                    }
                    catch (ThreadAbortException)
                    {
                        //aborted - return
                        return;
                    }
                    catch (WebException ex)
                    {
                        System.Diagnostics.Debug.WriteLine("=============: " + ex.Message);
                        // wait for a while before the next try
                        Thread.Sleep(250);
                    }
                    catch (ApplicationException ex)
                    {
                        System.Diagnostics.Debug.WriteLine("=============: " + ex.Message);
                        // wait for a while before the next try
                        Thread.Sleep(250);
                    }
                    catch (Exception ex)
                    {
                        System.Diagnostics.Debug.WriteLine("=============: " + ex.Message);
                    }
                }
            }
            finally
            {
                //cleanup and quit
                imageStream.Close();
                stopStream = false;
            }
        }
    }
}
