using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using MJPEGDecoder;
using System.Net;
using System.IO;
using System.Runtime.InteropServices;
using System.Threading;

namespace MacanumRobot
{
    public partial class WebcamViewer : Form
    {
        MJPEGDecodeAndDisplay decoder;

        public WebcamViewer()
        {
            InitializeComponent();
        }

        string address="", port="";
        private System.Windows.Forms.Timer timer;
        public WebcamViewer(String address, String port)
        {
            InitializeComponent();
            this.timer = new System.Windows.Forms.Timer();
            timer.Interval = 20;
            timer.Tick += new EventHandler(timer_Tick);
            timer.Start();
            decoder = new MJPEGDecodeAndDisplay(address + ":" + port + "/?action=stream");
            decoder.Picture += new MJPEGDecodeAndDisplay.PictureEventHandler(decoder_Picture);
            decoder.FPS += new MJPEGDecodeAndDisplay.FPSEventHandler(decoder_FPS);
            try
            {
                this.address = address;
                this.port = port;
                decoder.start();
            }
            catch (Exception ex)
            {
                MessageBox.Show("Error: " + ex.Message);
            }
        }

        void timer_Tick(object sender, EventArgs e)
        {
            KeyStateInfo left = KeyboardInfo.GetKeyState(Keys.Left);
            KeyStateInfo right = KeyboardInfo.GetKeyState(Keys.Right);
            KeyStateInfo up = KeyboardInfo.GetKeyState(Keys.Up);
            KeyStateInfo down = KeyboardInfo.GetKeyState(Keys.Down);

            if (left.IsPressed)
            {
                sendCommand("pan_plus");
            }

            if (right.IsPressed)
            {
                sendCommand("pan_minus");
            }

            if (up.IsPressed)
            {
                sendCommand("tilt_minus");
            }

            if (down.IsPressed)
            {
                sendCommand("tilt_plus");
            }

        }

        //can probably keep a stream open for commands
        public void sendCommand(string command)
        {
            sendCommand(command, "");
        }
        public void sendCommand(string command, string val)
        {
            HttpWebRequest req = null;
            HttpWebResponse resp = null;
            string commandAddress = "http://" + address + ":" + port + "/?action=command&command=" + command;
            if (val != "")
                commandAddress = commandAddress + "&value=" + val;
            Application.DoEvents();
            try
            {
                req = (HttpWebRequest)WebRequest.Create(commandAddress);
                resp = (HttpWebResponse)req.GetResponse();
                Stream dataStream = resp.GetResponseStream();
                StreamReader reader = new StreamReader(dataStream);
                string responseFromServer = reader.ReadToEnd();

                reader.Close();
                dataStream.Close();
                resp.Close();
            }
            catch
            {
            }
        }

        void decoder_FPS(object sender, MJPEGDecodeAndDisplay.FPSEventArgs e)
        {
            fpsLabel.Text = e.fps.ToString();
        }

        void decoder_Picture(object sender, MJPEGDecodeAndDisplay.PictureEventArgs e)
        {
            webcamView.Image = e.picture;
        }

        private void WebcamViewer_FormClosing(object sender, FormClosingEventArgs e)
        {
            stop();
        }

        public void stop()
        {
            decoder.stop();
        }

        private void webcamView_Resize(object sender, EventArgs e)
        {
            int height = (int)Math.Round(webcamView.Size.Width * 3 / 4.0);
            webcamView.Size = new Size(webcamView.Size.Width, height);
            this.MinimumSize = new System.Drawing.Size(this.MinimumSize.Width, height + 34); //34 for border/toolbar
        }

        private void resetPanTiltButton_Click(object sender, EventArgs e)
        {
            Thread resetThread = new Thread(new ThreadStart(resetWebcam));
            resetThread.Start();
        }

        private void resetWebcam()
        {
            sendCommand("reset_pan_tilt");
        }
    }
    public struct KeyStateInfo
    {
        private Keys key;
        private bool isPressed;
        private bool isToggled;

        public KeyStateInfo(Keys key, bool ispressed, bool istoggled)
        {
            this.key = key;
            isPressed = ispressed;
            isToggled = istoggled;
        }

        public static KeyStateInfo Default
        {
            get
            {
                return new KeyStateInfo(Keys.None, false, false);
            }
        }

        public Keys Key
        {
            get { return key; }
        }

        public bool IsPressed
        {
            get { return isPressed; }
        }

        public bool IsToggled
        {
            get { return isToggled; }
        }
    }
    public class KeyboardInfo
    {
        private KeyboardInfo() { }
        [DllImport("user32")]
        private static extern short GetKeyState(int vKey);

        public static KeyStateInfo GetKeyState(Keys key)
        {
            short keyState = GetKeyState((int)key);
            int low = Low(keyState), high = High(keyState);
            bool toggled = low == 1;
            bool pressed = high == 1;
            return new KeyStateInfo(key, pressed, toggled);
        }
        private static int High(int keyState)
        {
            return keyState > 0 ? keyState >> 0x10
                    : (keyState >> 0x10) & 0x1;
        }
        private static int Low(int keyState)
        {
            return keyState & 0xffff;
        }
    }
}